//
//  ImageColorAnalyzer.m
//  ColorArt
//
// Redistribution and use, with or without modification, are permitted provided that the following conditions are met:
//
// - Redistributions must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
// - Neither the name of Panic Inc nor the names of its contributors may be used to endorse or promote works derived from this software without specific prior written permission from Panic Inc.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PANIC INC BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Modified by David Schimon 1 / 2013

#import "ImageColorAnalyzer.h"

@interface PCCountedColor : NSObject

@property (assign) NSUInteger count;
@property (retain) UIColor *color;

- (id)initWithColor:(UIColor*)color count:(NSUInteger)count;

@end

@interface UIColor (DarkAdditions)

- (BOOL)pc_isDarkColor;
- (BOOL)pc_isDistinct:(UIColor*)compareColor;
- (UIColor*)pc_colorWithMinimumSaturation:(CGFloat)saturation;
- (BOOL)pc_isBlackOrWhite;
- (BOOL)pc_isContrastingColor:(UIColor*)color;

@end

////////////////////////////////////

@implementation ImageColorAnalyzer

#pragma mark - Initialize / Release

- (id)initWithImage:(UIImage *)image
{
    if (self = [super init]) {
        _image = image;
    }
    return self;
}

- (void)analyzeImage
{
	NSCountedSet *imageColors = nil;
	_backgroundColor = [self findEdgeColor:_image imageColors:&imageColors];
    UIColor *primaryColor = nil; UIColor *secondaryColor = nil; UIColor *detailColor = nil;
	[self findTextColors:imageColors primaryColor:&primaryColor secondaryColor:&secondaryColor detailColor:&detailColor backgroundColor:_backgroundColor];
    
    _primaryColor = primaryColor;
    _secondaryColor = secondaryColor;
    _detailColor = detailColor;
}

- (UIColor*)findEdgeColor:(UIImage*)image imageColors:(NSCountedSet**)colors
{
    CGImageRef imageRef = [image CGImage];
	
	NSInteger pixelsWide = CGImageGetWidth(imageRef);
	NSInteger pixelsHigh = CGImageGetHeight(imageRef);
    
	NSCountedSet *imageColors = [[NSCountedSet alloc] initWithCapacity:pixelsWide * pixelsHigh];
	NSCountedSet *leftEdgeColors = [[NSCountedSet alloc] initWithCapacity:pixelsHigh];
    
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    unsigned char *rawData = (unsigned char*) calloc(pixelsHigh * pixelsWide * 4, sizeof(unsigned char));
    NSUInteger bytesPerPixel = 4;
    NSUInteger bytesPerRow = bytesPerPixel * pixelsWide;
    NSUInteger bitsPerComponent = 8;
    CGContextRef context = CGBitmapContextCreate(rawData, pixelsWide, pixelsHigh,
                                                 bitsPerComponent, bytesPerRow, colorSpace,
                                                 kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big);
    CGColorSpaceRelease(colorSpace);
    
    CGContextDrawImage(context, CGRectMake(0, 0, pixelsHigh, pixelsWide), imageRef);
    CGContextRelease(context);
    
	for ( NSUInteger x = 0; x < pixelsWide; x++ )
	{
		for ( NSUInteger y = 0; y < pixelsHigh; y++ )
		{
            int byteIndex = (bytesPerRow * y) + x * bytesPerPixel;
            CGFloat red   = (rawData[byteIndex]     * 1.0) / 255.0;
            CGFloat green = (rawData[byteIndex + 1] * 1.0) / 255.0;
            CGFloat blue  = (rawData[byteIndex + 2] * 1.0) / 255.0;
            CGFloat alpha = (rawData[byteIndex + 3] * 1.0) / 255.0;
            UIColor *color = [UIColor colorWithRed:red green:green blue:blue alpha:alpha];
            
			if ( x == 0 )
			{
				[leftEdgeColors addObject:color];
			}
			
			[imageColors addObject:color];
		}
	}
    
    free(rawData);
    
	*colors = imageColors;
    
    
	NSEnumerator *enumerator = [leftEdgeColors objectEnumerator];
	UIColor *curColor = nil;
	NSMutableArray *sortedColors = [NSMutableArray arrayWithCapacity:[leftEdgeColors count]];
    
	while ( (curColor = [enumerator nextObject]) != nil )
	{
		NSUInteger colorCount = [leftEdgeColors countForObject:curColor];
        
		if ( colorCount <= 2 ) // prevent using random colors, threshold should be based on input image size
			continue;
		
		PCCountedColor *container = [[PCCountedColor alloc] initWithColor:curColor count:colorCount];
		
		[sortedColors addObject:container];
	}
    
	[sortedColors sortUsingSelector:@selector(compare:)];
	
    
	PCCountedColor *proposedEdgeColor = nil;
	
	if ( [sortedColors count] > 0 )
	{
		proposedEdgeColor = [sortedColors objectAtIndex:0];
        
		if ( [proposedEdgeColor.color pc_isBlackOrWhite] ) // want to choose color over black/white so we keep looking
		{
			for ( NSInteger i = 1; i < [sortedColors count]; i++ )
			{
				PCCountedColor *nextProposedColor = [sortedColors objectAtIndex:i];
				
				if (((double)nextProposedColor.count / (double)proposedEdgeColor.count) > .3 ) // make sure the second choice color is 30% as common as the first choice
				{
					if ( ![nextProposedColor.color pc_isBlackOrWhite] )
					{
						proposedEdgeColor = nextProposedColor;
						break;
					}
				}
				else
				{
					// reached color threshold less than 30% of the original proposed edge color so bail
					break;
				}
			}
		}
	}
	
	return proposedEdgeColor.color;
}


- (void)findTextColors:(NSCountedSet*)colors primaryColor:(UIColor**)primaryColor secondaryColor:(UIColor**)secondaryColor detailColor:(UIColor**)detailColor backgroundColor:(UIColor*)backgroundColor
{
	NSEnumerator *enumerator = [colors objectEnumerator];
	UIColor *curColor = nil;
	NSMutableArray *sortedColors = [NSMutableArray arrayWithCapacity:[colors count]];
	BOOL findDarkTextColor = ![backgroundColor pc_isDarkColor];
	
	while ( (curColor = [enumerator nextObject]) != nil )
	{
		curColor = [curColor pc_colorWithMinimumSaturation:.15]; // make sure color isn't too pale or washed out
        
		if ( [curColor pc_isDarkColor] == findDarkTextColor )
		{
			NSUInteger colorCount = [colors countForObject:curColor];
			
			//if ( colorCount <= 2 ) // prevent using random colors, threshold should be based on input image size
			//	continue;
			
			PCCountedColor *container = [[PCCountedColor alloc] initWithColor:curColor count:colorCount];
			
			[sortedColors addObject:container];
		}
	}
	
	[sortedColors sortUsingSelector:@selector(compare:)];
	
	for ( PCCountedColor *curContainer in sortedColors )
	{
		curColor = curContainer.color;
        
		if ( *primaryColor == nil )
		{
			if ( [curColor pc_isContrastingColor:backgroundColor] )
				*primaryColor = curColor;
		}
		else if ( *secondaryColor == nil )
		{
			if ( ![*primaryColor pc_isDistinct:curColor] || ![curColor pc_isContrastingColor:backgroundColor] )
				continue;
            
			*secondaryColor = curColor;
		}
		else if ( *detailColor == nil )
		{
			if ( ![*secondaryColor pc_isDistinct:curColor] || ![*primaryColor pc_isDistinct:curColor] || ![curColor pc_isContrastingColor:backgroundColor] )
				continue;
            
			*detailColor = curColor;
			break;
		}
	}
}

@end

@implementation PCCountedColor

- (id)initWithColor:(UIColor*)color count:(NSUInteger)count
{
	self = [super init];
	
	if ( self )
	{
		self.color = color;
		self.count = count;
	}
	
	return self;
}

- (NSComparisonResult)compare:(PCCountedColor*)object
{
	if ( [object isKindOfClass:[PCCountedColor class]] )
	{
		if ( self.count < object.count )
		{
			return NSOrderedDescending;
		}
		else if ( self.count == object.count )
		{
			return NSOrderedSame;
		}
	}
    
	return NSOrderedAscending;
}

@end

///////////////////////////////////////

@implementation UIColor (DarkAdditions)

- (BOOL)pc_isDarkColor
{
	CGFloat r, g, b, a;
    
	[self getRed:&r green:&g blue:&b alpha:&a];
	
	CGFloat lum = 0.2126 * r + 0.7152 * g + 0.0722 * b;
    
	if ( lum < .5 )
	{
		return YES;
	}
	
	return NO;
}

- (BOOL)pc_isDistinct:(UIColor*)compareColor
{
	CGFloat r, g, b, a;
	CGFloat r1, g1, b1, a1;
    
	[self getRed:&r green:&g blue:&b alpha:&a];
	[compareColor getRed:&r1 green:&g1 blue:&b1 alpha:&a1];
    
	CGFloat threshold = .25; //.15
    
	if ( fabs(r - r1) > threshold ||
		fabs(g - g1) > threshold ||
		fabs(b - b1) > threshold ||
		fabs(a - a1) > threshold )
    {
        // check for grays, prevent multiple gray colors
        
        if ( fabs(r - g) < .03 && fabs(r - b) < .03 )
        {
            if ( fabs(r1 - g1) < .03 && fabs(r1 - b1) < .03 )
                return NO;
        }
        
        return YES;
    }
    
	return NO;
}

- (UIColor*)pc_colorWithMinimumSaturation:(CGFloat)minSaturation
{
    CGFloat hue = 0.0;
    CGFloat saturation = 0.0;
    CGFloat brightness = 0.0;
    CGFloat alpha = 0.0;
    
    [self getHue:&hue saturation:&saturation brightness:&brightness alpha:&alpha];
    
    if ( saturation < minSaturation ) {
        return [UIColor colorWithHue:hue saturation:minSaturation brightness:brightness alpha:alpha];
    }
	
	return self;
}

- (BOOL)pc_isBlackOrWhite
{
    CGFloat r, g, b, a;
    
    [self getRed:&r green:&g blue:&b alpha:&a];
    
    if ( r > .91 && g > .91 && b > .91 )
        return YES; // white
    
    if ( r < .09 && g < .09 && b < .09 )
        return YES; // black
	
	return NO;
}

- (BOOL)pc_isContrastingColor:(UIColor*)color
{
	
	if (color != nil)
	{
		CGFloat br, bg, bb, ba;
		CGFloat fr, fg, fb, fa;
		
		[self getRed:&br green:&bg blue:&bb alpha:&ba];
		[color getRed:&fr green:&fg blue:&fb alpha:&fa];
        
		CGFloat bLum = 0.2126 * br + 0.7152 * bg + 0.0722 * bb;
		CGFloat fLum = 0.2126 * fr + 0.7152 * fg + 0.0722 * fb;
        
		CGFloat contrast = 0.;
		
		if ( bLum > fLum )
			contrast = (bLum + 0.05) / (fLum + 0.05);
		else
			contrast = (fLum + 0.05) / (bLum + 0.05);
        
		//return contrast > 3.0; //3-4.5 W3C recommends a minimum ratio of 3:1
		return contrast > 1.6;
	}
	
	return YES;
}

@end
